package jadex.bridge;

import jadex.bridge.service.annotation.Reference;
import jadex.commons.SUtil;

import java.net.URL;

/**
 *  Default implementation for resource identification.
 */
@Reference(local=true, remote=false)
public class LocalResourceIdentifier implements ILocalResourceIdentifier
{
	//-------- attributes --------
	
	/** The component identifier. */
	protected IComponentIdentifier cid;
	
	/** The URL. */
	protected URL url;
	
	/** The host id. */
	protected String hostid;
	
	//-------- constructors --------

	/**
	 *  Create a resource identifier.
	 */
	public LocalResourceIdentifier()
	{
		// bean constructor
	}
	
	/**
	 *  Create a resource identifier.
	 *  @param cid The platform identifier.
	 *  @param url The local URL.
	 */
	public LocalResourceIdentifier(IComponentIdentifier cid, URL url)
	{
		this(cid, url, SUtil.getMacAddress());
	}
		
	/**
	 *  Create a resource identifier.
	 *  @param cid The platform identifier.
	 *  @param url The local URL.
	 */
	public LocalResourceIdentifier(IComponentIdentifier cid, URL url, String hostid)
	{
		if(cid==null)
		{
			throw new IllegalArgumentException("Cid must not null.");
		}
		if(url==null)
		{
			throw new IllegalArgumentException("Url must not null.");
		}
		if(hostid==null)
		{
			hostid = cid.getName(); // in case no mac is available use cid (strict as before)
		}
		if(url.toString().indexOf("..")!=-1)
		{
			throw new IllegalArgumentException("Url must use canonical path: "+url);
		}
		if(url.getFile().startsWith("."))
		{
			throw new IllegalArgumentException("Url must be absolute: "+url);
		}
		
		this.hostid = hostid;
		this.cid = cid;
		this.url = url;
	}

	//-------- methods --------
	
	/**
	 *  Get the platform identifier belonging to the resource identifier.
	 *  @return The component identifier of the platform. 
	 */
	public IComponentIdentifier	getComponentIdentifier()
	{
		return cid;
	}
	
	/**
	 *  Get the url.
	 *  @return The resource url.
	 */
	public URL	getUrl()
	{
		return url;
	}
	
	/**
	 *  Get the host identifier.
	 *  @return The host identifier.
	 */
	public String	getHostIdentifier()
	{
		return hostid;
	}
	
	/**
	 *  Set the platform identifier belonging to the resource identifier.
	 *  @param cid The component identifier of the platform. 
	 */
	public void	setComponentIdentifier(IComponentIdentifier cid)
	{
		this.cid	= cid;
	}
	
	/**
	 *  Set the url.
	 *  @param url The resource url.
	 */
	public void	setUrl(URL url)
	{
		this.url	= url;
	}
	
	/**
	 *  Set the host identifier.
	 *  @param hostid The host identifier.
	 */
	public void	setHostIdentifier(String hostid)
	{
		this.hostid = hostid;
	}
	
	/**
	 *  Get the hashcode.
	 */
	public int hashCode()
	{
		final int prime = 31;
		int result = 1;
//		result = prime * result + (cid!=null? cid.hashCode(): 0);
		result = prime * result + hostid.hashCode();
		result = prime * result + url.hashCode();
		return result;
	}

	/**
	 *  Test if equals.

	 */
	public boolean equals(Object obj)
	{
		boolean ret = false;
		if(obj instanceof ILocalResourceIdentifier)
		{
			ILocalResourceIdentifier other = (ILocalResourceIdentifier)obj;
//			ret = SUtil.equals(getComponentIdentifier(), other.getComponentIdentifier())
//				&& SUtil.equals(getUrl(), other.getUrl());
			ret = SUtil.equals(getHostIdentifier(), other.getHostIdentifier())
				&& SUtil.equals(getUrl(), other.getUrl());
		}
		return ret;
	}
	
	/**
	 *  Get a string representation of this object.
	 */
	public String	toString()
	{
		return url+"-"+hostid+"@"+cid;
	}
}
